"""Data object for language options.

Program strings that vary with the selected language.

Attributes:
    verse_markers: The equivalent of 'v.' and vv.' in English.
    number_prefixes: The equivalent of 'I' and 'First' (etc) in English.
"""

from typing import Dict, List

from pydantic import BaseModel


class Language(BaseModel):
    verse_markers: List[str]
    ambiguous_aliases: List[str]
    number_prefixes: Dict[str, List[str]]



from context import *

from refspy.indexers import index_book_aliases, index_books
from refspy.libraries.en_US import NT
from refspy.navigator import Navigator
from refspy.reference import chapter_reference as ch

books = index_books([NT])
book_aliases = index_book_aliases([NT])
__ = Navigator(books, book_aliases)


def test_prev_chapter():
    assert __.prev_chapter(ch(NT.id, 6, 7)) == ch(NT.id, 6, 6)  # Rom 7 -> Rom 6
    assert __.prev_chapter(ch(NT.id, 6, 1)) == ch(NT.id, 5, 28)  # Rom 1 -> Acts 28
    assert __.prev_chapter(ch(NT.id, 1, 1)) is None  # Matt 1 -> None


def test_next_chapter():
    assert __.next_chapter(ch(NT.id, 6, 7)) == ch(NT.id, 6, 8)  # Rom 7 -> Rom 8
    assert __.next_chapter(ch(NT.id, 6, 16)) == ch(NT.id, 7, 1)  # Rom 16 -> 1 Cor 1
    assert __.next_chapter(ch(NT.id, 27, 22)) is None  # Rev 22 -> None



"""Initialisation helpers for Refspy package."""

from typing import List
from refspy.config import LANGUAGES, LIBRARIES
from refspy.language import Language
from refspy.library import Library


def get_language(language_name: str) -> Language:
    """Lookup key in `refspy.config.LANGUAGES` dict."""
    if language_name in LANGUAGES:
        return LANGUAGES[language_name]
    else:
        raise ValueError(f"Language '{language_name}' not found.")


def get_canon(canon_name: str, locale_name: str) -> List[Library]:
    """Lookup keys in `refspy.config.LIBRARIES` dict."""
    if canon_name in LIBRARIES:
        if locale_name in LIBRARIES[canon_name]:
            return LIBRARIES[canon_name][locale_name]
    raise ValueError(f"Canon '{canon_name}' not found for locale '{locale_name}'.")



from context import *

import pytest

from refspy import refspy
from refspy.languages.english import ENGLISH
from refspy.utils import (
    add_space_after_book_number,
    get_unnumbered_book_aliases,
    normalize_spacing,
    parse_number,
    sequential_replace,
    sequential_replace_tuples,
    strip_book_number,
    strip_space_after_book_number,
    url_param,
    url_escape,
)


__ = refspy()

def test_parse_number():
    text = "USD $50"
    assert parse_number(text) == 50


def test_parse_number_raises_value_error():
    text = "My offer is this: nothing."
    with pytest.raises(ValueError):
        assert parse_number(text) == 50


def test_url_param():
    assert url_param("1 Cor 3:4–5") == "1+cor+3.4-5"

def test_url_escape():
    assert url_escape("1 Cor 3:4–5") == "1%20Cor%203%3A4-5"


def test_strip_book_number():
    assert strip_book_number('2 Tim') == 'Tim'
    assert strip_book_number('2Tim') == '2Tim'
    assert strip_book_number('1st') == '1st'


def test_strip_space_after_book_number():
    assert strip_space_after_book_number('2 Tim') == '2Tim'
    assert strip_space_after_book_number('2  Tim') == '2 Tim'
    assert strip_space_after_book_number('2Tim') == '2Tim'
    assert strip_space_after_book_number('1st') == '1st'


def test_add_space_after_book_number():
    """ Remove space between any leading digit and all subsequent text.

    e.g. '2Tim' becomes '2 Tim'.
    """
    unnumbered_book_aliases = get_unnumbered_book_aliases(__.book_aliases)

    assert add_space_after_book_number('2Tim', unnumbered_book_aliases, ENGLISH.number_prefixes) == '2 Tim'
    assert add_space_after_book_number('IITim', unnumbered_book_aliases, ENGLISH.number_prefixes) == '2 Tim'
    assert add_space_after_book_number('2ndTim', unnumbered_book_aliases, ENGLISH.number_prefixes) == '2 Tim'
    assert add_space_after_book_number('SecondTim', unnumbered_book_aliases, ENGLISH.number_prefixes) == '2 Tim'
    assert add_space_after_book_number('2 Tim', unnumbered_book_aliases, ENGLISH.number_prefixes) == '2 Tim'
    assert add_space_after_book_number('II Tim', unnumbered_book_aliases, ENGLISH.number_prefixes) == 'II Tim'
    assert add_space_after_book_number('2nd Tim', unnumbered_book_aliases, ENGLISH.number_prefixes) == '2nd Tim'
    assert add_space_after_book_number('Second Tim', unnumbered_book_aliases, ENGLISH.number_prefixes) == 'Second Tim'


def test_normalize_spacing():
    text = "  a  \t\t b    c  \n\r  d   e   "
    assert normalize_spacing(text) == " a b c d e "


def test_sequential_replace_tuples():
    text = "1 Cor 1; 1 Cor 1:1; 1 Cor 1:1-4"
    tuples = [
        ("1 Cor 1", "XXXXXXX"),
        ("1 Cor 1:1", "YYYYYYYYY"),
        ("1 Cor 1:1-4", "ZZZZZZZZZZZ"),
    ]
    result = sequential_replace_tuples(text, tuples)
    assert result == "XXXXXXX; YYYYYYYYY; ZZZZZZZZZZZ"


def test_sequential_replace():
    text = "1 Cor 1; 1 Cor 1:1; 1 Cor 1:1-4"
    strs = ["1 Cor 1", "1 Cor 1:1", "1 Cor 1:1-4"]
    tags = ["XXXXXXX", "YYYYYYYYY", "ZZZZZZZZZZZ"]
    result = sequential_replace(text, strs, tags)
    assert result == "XXXXXXX; YYYYYYYYY; ZZZZZZZZZZZ"



from context import *

from refspy import refspy

def test_include_two_letter_aliases():
    __ = refspy()  # <-- include two-letter aliases by default
 
    match, ref = __.first_reference("1Ti 2")
    assert match == "1Ti 2"
    assert ref == __.bcv('1 Tim', 2)
    match, ref = __.first_reference("Ge 1:1")
    assert match == "Ge 1:1"
    assert ref == __.bcv('Gen', 1, 1)
    match, ref = __.first_reference("Jn 1:1")
    assert match == "Jn 1:1"
    assert ref == __.bcv('John', 1, 1)


def test_exclude_two_letter_aliases():
    __ = refspy(include_two_letter_aliases=False)  # <-- but not ambiguous aliases

    match, ref = __.first_reference("1Ti 2")
    assert match == None
    assert ref == None
    match, ref = __.first_reference("Ob 2")
    assert match == None
    assert ref == None
    match, ref = __.first_reference("Am 1:1")
    assert match == None
    assert ref == None
    match, ref = __.first_reference("Is 1:1")
    assert match == None
    assert ref == None


def test_include_ambiguous_aliases():
    __ = refspy(include_ambiguous_aliases=True)

    match, ref = __.first_reference("1Ti 2")
    assert match == "1Ti 2"
    assert ref == __.bcv('1 Tim', 2)
    match, ref = __.first_reference("Ob 2:1")
    assert match == "Ob 2:1"
    assert ref == __.bcv('Obadiah', 2, 1)
    match, ref = __.first_reference("Am 2")
    assert match == "Am 2"
    assert ref == __.bcv('Amos', 2)
    match, ref = __.first_reference("Is 1:1")
    assert match == "Is 1:1"
    assert ref == __.bcv('Isaiah', 1, 1)


def test_not_matching_parts_of_words():
    __ = refspy(include_two_letter_aliases=True, include_ambiguous_aliases=True)

    match, ref = __.first_reference("Test test test. Especially the last part.")
    #                                          ^^  Match Esther?
    assert match == None
    assert ref == None




"""Data object for a book.

Books are referenced by their names, abbrevs (short names), and aliases.

Attributes:
    id: 13
    name: '1 Thesslonians'
    abbrev: '1 Thess'
    aliases: ['1 Th'],
    chapters: 5

Note:
    Aliases should be capitalised.

If the book has a number (e.g. '1' Corinthians), then the
`refspy.language.Language` object's `number_prefixes` property defines any
substitute prefixes like 'I' or 'First'.

A book with only one chapter will be formatted without chapter numbers.
"""

from typing import List
from pydantic import BaseModel

from refspy.number import Number


class Book(BaseModel):
    id: Number
    name: str
    abbrev: str
    aliases: List[str]
    chapters: int



"""Navigation functions for Reference objects"""

from typing import Dict, Tuple

from refspy.book import Book
from refspy.range import range
from refspy.reference import Reference, reference
from refspy.verse import Number, verse


class Navigator:
    def __init__(
        self,
        books: Dict[Tuple[Number, Number], Book],
        book_aliases: Dict[str, Tuple[Number, Number]],
    ) -> None:
        """
        Note:
            Taking prev and next from references makes the most sense if ranges
            are sorted or have been collated by book and chapter.
        Example:

            Using the Navigator module on its own:

            ```
            from refspy.indexers import index_book_aliases, index_books
            from refspy.libraries.en_US import NT, OT
            from refspy.navigator import Navigator
            from refspy.reference import chapter_reference

            libraries = [OT, NT]
            navigator = Navigator(
                index_books(libraries),
                index_book_aliases(libraries)
            )
            matt_1 = chapter_reference(NT.id, 1, 1)
            matt_2 = chapter_reference(NT.id, 1, 2)
            assert navigator.next_chapter(matt_1) == matt_2
            assert navigator.prev_chapter(matt_1) is None
            ```

        Example:

            Using the navigator within the `refspy` package.

            ```
            from refspy import refspy

            __ = refspy()
            matt_1 = __.bcv('Matt', 1)
            matt_2 = __.bcv('Matt', 2)
            assert __.next_chapter(matt_1) == matt_2
            assert __.prev_chapter(matt_1) is None
            ```
        """

        self.books: Dict[Tuple[Number, Number], Book] = books
        """Lookup `refspy.book.Book` by `(library.id, book.id)`."""

        self.book_aliases: Dict[str, Tuple[Number, Number]] = book_aliases
        """Lookup `(library.id, book.id)` by book alias."""

    def prev_chapter(self, ref: Reference) -> Reference | None:
        """Find a reference to the previous chapter.

        We step into preceding books but not preceding libraries.

        Returns:
            None: If this is the first chapter of a `refspy.library.Library`.
        """
        v1 = ref.ranges[0].start
        if v1.chapter > 1:
            return reference(
                range(
                    verse(v1.library, v1.book, v1.chapter - 1, 1),
                    verse(v1.library, v1.book, v1.chapter - 1, 999),
                )
            )
        else:
            if (v1.library, v1.book - 1) not in self.books:
                return None
            prev_book = self.books[v1.library, v1.book - 1]
            end_chapter = prev_book.chapters
            return reference(
                range(
                    verse(v1.library, prev_book.id, end_chapter, 1),
                    verse(v1.library, prev_book.id, end_chapter, 999),
                )
            )

    def next_chapter(self, ref: Reference) -> Reference | None:
        """Find a reference to the next chapter.

        We step into succeeding books but not succeeding libraries.

        Returns:
            None: if this is the last chapter of a `refspy.library.Library`.
        """
        v1 = ref.ranges[0].start
        end_chapter = self.books[v1.library, v1.book].chapters
        if v1.chapter < end_chapter:
            return reference(
                range(
                    verse(v1.library, v1.book, v1.chapter + 1, 1),
                    verse(v1.library, v1.book, v1.chapter + 1, 999),
                )
            )
        else:
            if (v1.library, v1.book + 1) not in self.books:
                return None
            next_book = self.books[v1.library, v1.book + 1]
            return reference(
                range(
                    verse(v1.library, next_book.id, 1, 1),
                    verse(v1.library, next_book.id, 1, 999),
                )
            )



import pytest
from context import *

from refspy import refspy
from refspy.book import Book
from refspy.languages.english import ENGLISH
from refspy.libraries.en_US import NT
from refspy.library import Library
from refspy.manager import Manager
from refspy.range import range
from refspy.reference import reference, verse_reference
from refspy.verse import verse

BOOK = Book(id=2, name="Book", abbrev="Bk", aliases=["vol"], chapters=3)

LIBRARY = Library(id=1, name="Library", abbrev="Lib", books=[BOOK])

REFERENCES = [
    # 'Book 1:1–2' in Library
    reference(range(verse(LIBRARY.id, BOOK.id, 1, 1), verse(LIBRARY.id, BOOK.id, 1, 2)))
]

__ = Manager([LIBRARY], ENGLISH)
REFSPY = refspy()


def test_init():
    assert __.libraries[LIBRARY.id] == LIBRARY
    assert __.book_aliases["Book"] == (LIBRARY.id, BOOK.id)


def test_make_index():
    tuples = __.find_references("Book 1:2, 2:1, 3:4, 1:4-5, 7, 3:6")
    text = __.make_index([ref for _, ref in tuples if ref])
    assert text == "Bk 1:2, 4–5, 7; 2:1; 3:4, 6"


def test_make_hotspots():
    tuples = __.find_references("Book 1:2, 2:1, 3:4, 1:4-5, 7, 3:6")
    text = __.make_hotspots_text(
        [ref for _, ref in tuples if ref], max_chapters=2, min_references=2
    )
    assert text is not None
    assert "Bk 1, Bk 3" in text


def test_make_hotspots_lots():
    tuples = __.find_references("Book 1:2, 2:1, 3:4, 1:4-5, 7, 3:6, " * 100)
    text = __.make_hotspots_text(
        [ref for _, ref in tuples if ref], max_chapters=2, min_references=2
    )
    assert text is not None
    assert "Bk 1, Bk 3" in text


def test_make_hotspots_empty():
    text = __.make_hotspots_text([], max_chapters=2, min_references=2)
    assert text is None


def test_non_unique():
    """
    Repeating any name/alias values raise an error.
    """
    with pytest.raises(ValueError):
        _ = Manager([LIBRARY, LIBRARY], ENGLISH)


def test_sort_references():
    ref_1 = verse_reference(NT.id, 1, 2, 3)
    ref_2 = verse_reference(NT.id, 1, 2, 4)
    ref_3 = verse_reference(NT.id, 1, 2, 5)
    sorted_ref = __.sort_references([ref_3, ref_2, ref_1])
    assert sorted_ref.ranges == [ref_1.ranges[0], ref_2.ranges[0], ref_3.ranges[0]]


def test_merge_references():
    ref_1 = verse_reference(NT.id, 1, 2, 3)
    ref_2 = verse_reference(NT.id, 1, 2, 3, 4)
    ref_3 = verse_reference(NT.id, 1, 2, 4, 5)
    merged = __.merge_references([ref_3, ref_2, ref_1])
    assert len(merged.ranges) == 1
    assert merged == verse_reference(NT.id, 1, 2, 3, 5)


def test_combine_references():
    ref_1 = verse_reference(NT.id, 1, 2, 3)
    ref_2 = verse_reference(NT.id, 1, 2, 4)
    ref_3 = verse_reference(NT.id, 1, 2, 5, 6)
    combined = __.combine_references([ref_3, ref_2, ref_1])
    assert len(combined.ranges) == 1
    assert combined == verse_reference(NT.id, 1, 2, 3, 6)


def test_collate_by_id():
    collation = __.collate_by_id(REFERENCES)
    for library_id, book_collation in collation.items():
        assert library_id == LIBRARY.id
        for book_id, book_references in book_collation.items():
            assert book_id == BOOK.id
            assert book_references == REFERENCES


def test_collate():
    collation = __.collate(REFERENCES)
    for library, book_collation in collation:
        assert library == LIBRARY
        for book, book_references in book_collation:
            assert book == BOOK
            assert book_references == REFERENCES


def test_first_reference():
    text = "Book 1:1"
    _, reference = __.first_reference(text)
    assert reference is not None
    assert __.name(reference) == text


def test_find_references():
    tuples = __.find_references("Book 1:1–2:4")
    _, reference = tuples[0]
    assert reference is not None
    assert __.name(reference) == "Book 1:1–2:4"


def test_find_all():
    text = "Book 1:1, Book 2:2"
    tuples = list(__.find_references(text))
    assert tuples[0] == ("Book 1:1", __.r("Book 1:1"))
    assert tuples[1] == ("Book 2:2", __.r("Book 2:2"))


def test_next_chapter():
    __ = Manager([LIBRARY], ENGLISH)
    ch_1 = __.bcv("Book", 1)
    ch_2 = __.bcv("Book", 2)
    assert __.next_chapter(ch_1) == ch_2


def test_prev_chapter():
    __ = Manager([LIBRARY], ENGLISH)
    ch_2 = __.bcv("Book", 2)
    ch_1 = __.bcv("Book", 1)
    assert __.prev_chapter(ch_2) == ch_1


def test_bcv():
    bk = __.bcv("Book")
    assert __.name(bk) == "Book"
    bk_2 = __.bcv("Book", 2)
    assert __.name(bk_2) == "Book 2"
    bk_2_1 = __.bcv("Book", 2, 1)
    assert __.name(bk_2_1) == "Book 2:1"


def test_bcr():
    bk_2_135 = __.bcr("Book", 2, [1, 3, 5])
    assert __.name(bk_2_135) == "Book 2:1, 3, 5"
    bk_2_1357 = __.bcr("Book", 2, [(1, 3), (5, 7)])
    assert __.name(bk_2_1357) == "Book 2:1–3, 5–7"


def test_r():
    ref = __.r("Book 1:1")
    assert ref is not None
    assert __.name(ref) == "Book 1:1"
    ref = __.r("Book 1:1, 3")
    assert ref is not None
    assert __.name(ref) == "Book 1:1, 3"
    ref = __.r("Book 1:1-3")
    assert ref is not None
    assert __.name(ref) == "Book 1:1–3"
    ref = __.r("Book 1:1-3, 5-7")
    assert ref is not None
    assert __.name(ref) == "Book 1:1–3, 5–7"
    ref = __.r("Book 1:1–2:4")
    assert ref is not None
    assert __.name(ref) == "Book 1:1–2:4"
    ref = __.r("Book 1:1-2:4")
    assert ref is not None
    assert __.name(ref) == "Book 1:1–2:4"


def test_get_book():
    ref = __.r("Book 1:1")
    assert ref is not None
    book = __.get_book(ref)
    assert book.name == "Book"


def test_book_reference():
    ref = __.r("Book 1:1")
    assert ref is not None
    book_ref = __.book_reference(ref)
    assert __.name(book_ref) == "Book"


def test_chapter_reference():
    ref = __.r("Book 1:1")
    assert ref is not None
    chapter_ref = __.chapter_reference(ref)
    assert __.name(chapter_ref) == "Book 1"


def test_name():
    ref = __.r("Book 1:1")
    assert ref is not None
    assert __.name(ref) == "Book 1:1"


def test_book():
    ref = __.r("Book 1:1")
    assert ref is not None
    assert __.book(ref) == "Book"


def test_numbers():
    ref = __.r("Book 1:1")
    assert ref is not None
    assert __.numbers(ref) == "1:1"


def test_abbrev_name():
    ref = __.r("Book 1:1")
    assert ref is not None
    assert __.abbrev_name(ref) == "Bk 1:1"


def test_abbrev_book():
    ref = __.r("Book 1:1")
    assert ref is not None
    assert __.abbrev_book(ref) == "Bk"


def test_template():
    ref = __.r("1 Cor 2:3-4, 5")
    if ref is not None:
        assert __.template(ref, "x {NAME}") == "x 1 Corinthians 2:3–4, 5"
        assert __.template(ref, "x {BOOK}") == "x 1 Corinthians"
        assert __.template(ref, "x {NUMBERS}") == "x 2:3–4, 5"
        assert __.template(ref, "x {ABBREV_NAME}") == "x 1 Cor 2:3–4, 5"
        assert __.template(ref, "x {ABBREV_BOOK}") == "x 1 Cor"
        assert __.template(ref, "x {ABBREV_NUMBERS}") == "x 2:3–4, 5"
        assert __.template(ref, "x {ESC_NAME}") == "x 1%20Corinthians%202%3A3-4,%205"
        assert __.template(ref, "x {ESC_BOOK}") == "x 1%20Corinthians"
        assert __.template(ref, "x {ESC_NUMBERS}") == "x 2%3A3-4,%205"
        assert __.template(ref, "x {ESC_ABBREV_NAME}") == "x 1%20Cor%202%3A3-4,%205"
        assert __.template(ref, "x {ESC_ABBREV_BOOK}") == "x 1%20Cor"
        assert __.template(ref, "x {ESC_ABBREV_NUMBERS}") == "x 2%3A3-4,%205"
        assert __.template(ref, "x {PARAM_NAME}") == "x 1cor+2:3-4,+5"
        assert __.template(ref, "x {PARAM_BOOK}") == "x 1cor"
        assert __.template(ref, "x {PARAM_NUMBERS}") == "x 2:3-4,+5"



"""
Data object for reference formatting options.

The Format objects are used by the `refspy.formatter.Formatter` to turn
references into different kinds of strings. Properties are named by their
characters in normal English formatting.

Attributes:
    book_only: False
    colon: ':'
    comma: ','
    dash: '-'
    number_only: False
    property: 'name'
    semicolon: '; '
    space: ' '

Note:
    If book_only is true, number_only is ignored.
"""

from pydantic import BaseModel


class Format(BaseModel):
    book_only: bool
    colon: str
    comma: str
    dash: str
    number_only: bool
    dash: str
    property: str | None
    semicolon: str
    space: str


NAME_FORMAT = Format(
    colon=":",
    comma=", ",
    dash="–",
    book_only=False,
    number_only=False,
    property="name",
    semicolon="; ",
    space=" ",
)
BOOK_FORMAT = Format(
    colon=":",
    comma=", ",
    dash="–",
    book_only=True,
    number_only=False,
    property="name",
    semicolon="; ",
    space=" ",
)
NUMBER_FORMAT = Format(
    colon=":",
    comma=", ",
    dash="–",
    book_only=False,
    number_only=True,
    property=None,
    semicolon="; ",
    space=" ",
)
ABBREV_NAME_FORMAT = Format(
    colon=":",
    comma=", ",
    dash="–",
    book_only=False,
    number_only=False,
    property="abbrev",
    semicolon="; ",
    space=" ",
)
ABBREV_BOOK_FORMAT = Format(
    colon=":",
    comma=", ",
    dash="–",
    book_only=True,
    number_only=False,
    property="abbrev",
    semicolon="; ",
    space=" ",
)
ABBREV_NUMBER_FORMAT = Format(
    colon=":",
    comma=", ",
    dash="–",
    book_only=False,
    number_only=True,
    property="abbrev",
    semicolon="; ",
    space=" ",
)



from context import *

from refspy.formatter import Formatter
from refspy.format import ABBREV_NAME_FORMAT, NAME_FORMAT, NUMBER_FORMAT
from refspy.indexers import index_book_aliases, index_books
from refspy.range import range
from refspy.reference import reference
from refspy.verse import verse

from tests.data import TEST_LIBRARY, TEST_LIBRARY_2

books = index_books([TEST_LIBRARY, TEST_LIBRARY_2])
book_aliases = index_book_aliases([TEST_LIBRARY, TEST_LIBRARY_2])
fmt = Formatter(books, book_aliases)


def test_book_reference():
    ref = reference(range(verse(1, 2, 1, 1), verse(1, 2, 999, 999)))

    assert fmt.format(ref, NAME_FORMAT) == "Big Book"


def test_book_range():
    ref = reference(range(verse(1, 2, 1, 1), verse(1, 3, 999, 999)))

    assert fmt.format(ref, ABBREV_NAME_FORMAT) == "Big–Small"


def test_chapter_reference():
    ref = reference(range(verse(1, 2, 3, 1), verse(1, 2, 3, 999)))

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 3"


def test_chapter_range():
    ref = reference(range(verse(1, 2, 3, 1), verse(1, 2, 4, 999)))

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 3–4"


def test_single_range():
    ref = reference(range(verse(1, 2, 1, 1), verse(1, 2, 1, 2)))

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:1–2"
    assert fmt.format(ref, ABBREV_NAME_FORMAT) == "Big 1:1–2"
    assert fmt.format(ref, NUMBER_FORMAT) == "1:1–2"


def test_multiple_verse_range():
    ref = reference(
        range(verse(1, 2, 1, 1), verse(1, 2, 1, 2)),
        range(verse(1, 2, 1, 7), verse(1, 2, 1, 9)),
    )

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:1–2, 7–9"


def test_inter_chapter_ranges():
    ref = reference(
        range(verse(1, 2, 1, 20), verse(1, 2, 2, 10)),
        range(verse(1, 2, 3, 21), verse(1, 2, 4, 11)),
    )

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:20–2:10; 3:21–4:11"


def test_multiple_chapter_range():
    ref = reference(
        range(verse(1, 2, 1, 1), verse(1, 2, 1, 2)),
        range(verse(1, 2, 2, 1), verse(1, 2, 2, 2)),
    )

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:1–2; 2:1–2"


def test_multiple_book_range():
    ref = reference(
        range(verse(1, 2, 1, 1), verse(1, 2, 1, 2)),
        range(verse(1, 3, 1, 1), verse(1, 3, 1, 2)),
    )

    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:1–2; Small Book 1–2"


def test_multiple_library_range():
    ref = reference(
        range(verse(1, 2, 1, 1), verse(1, 2, 1, 2)),
        range(verse(2, 1, 2, 1), verse(2, 1, 2, 1)),
    )
    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:1–2; 4 Book 2:1"


def test_book_chapters_greater_than_1():
    """
    Book ID 2 is Big Book; references should show chapter numbers.
    """
    ref = reference(
        range(verse(1, 2, 1, 1), verse(1, 2, 1, 2)),
    )
    assert fmt.format(ref, NAME_FORMAT) == "Big Book 1:1–2"


def test_book_chapter_equal_1():
    """
    Book ID 3 is Small Book; references should not show chapter numbers.
    """
    ref = reference(
        range(verse(1, 3, 1, 1), verse(1, 3, 1, 2)),
    )
    assert fmt.format(ref, NAME_FORMAT) == "Small Book 1–2"



"""Data type for index numbers.

An index number can be used to store a verse in a database using an `UNSIGNED
INT(12)` field.

It is created by multiplying each constituent number (library, book, chapter,
verse) by powers of 1000.

Example:
    `verse(1, 2, 3, 4)` becomes `1002003004`
"""

from typing import Annotated
from annotated_types import Ge, Le  # >=, <=

Index = Annotated[int, Ge(1001001001), Le(999999999999)]



"""Configure libraries and languages."""

from typing import Dict, List

from refspy.language import Language
from refspy.languages.english import ENGLISH
from refspy.libraries.en_US import DC, DC_ORTHODOX, NT, OT
from refspy.library import Library

LIBRARIES: Dict[str, Dict[str, List[Library]]] = {
    "protestant": {"en_US": [OT, NT]},
    "catholic": {"en_US": [OT, DC, NT]},
    "orthodox": {"en_US": [OT, DC, DC_ORTHODOX, NT]},
}
"""A dictionary of available locales for available libraries, used for
shorthand library invocation.

Example:
    ```
    from refspy.libraries.en_US import NT, OT

    LIBRARIES = { "protestant": {"en_US": [OT, NT] } }
    ```
"""

LANGUAGES: Dict[str, Language] = {
    "en": ENGLISH,
}
"""A dictionary of available languages, used for shorthand library invocation.

Language names use the first two chars of locale names, e.g. 'en_US' is 'en'.

Example:
    ```
    from refspy.languages.english import ENGLISH

    LANGUAGES = { "en": ENGLISH }
    ```
"""



"""Indexing functions for libraries, books, and book aliases."""

from typing import Dict, List, Tuple
from refspy import language
from refspy.book import Book
from refspy.library import Library
from refspy.utils import strip_book_number, strip_space_after_book_number, url_param
from refspy.verse import Number

PARAM_CHAR_LIMIT = 20
"""The maximum book name or alias length to convert to URL parameter names."""


def index_libraries(libraries: List[Library]) -> Dict[Number, Library]:
    """Create a lookup dict for libraries by library.id"""
    index = dict()
    for library in libraries:
        index[library.id] = library
    return index


def index_books(libraries: List[Library]) -> Dict[Tuple[Number, Number], Book]:
    """Create a lookup dict for books by (library.id, book.id)"""
    index = dict()
    for library in libraries:
        for book in library.books:
            index[library.id, book.id] = book
    return index


def add_unique_book_alias(
    index: Dict[str, Tuple[Number, Number]],
    alias: str,
    library_id: Number,
    book_id: Number,
    strict: bool = True,
):
    """Add an index entry for a library and book if none exists

    Modifies the passed index dict.

    Args:
        strict: raise an error in the case of a collision between
            alias values.
    """
    if alias in index and strict:
        raise ValueError(f"Book alias '{alias}' is not unique.")
    if alias == "":
        raise ValueError(f"Book alias is '' for ({library_id},{book_id})")
    else:
        index[alias] = (library_id, book_id)


def index_book_aliases(
    libraries: List[Library],
    include_two_letter_aliases: bool = True,
    ignore_aliases: List[str] = [],
    strict: bool = True
) -> Dict[str, Tuple[Number, Number]]:
    """Create a lookup table for library and book IDs by book aliases.

    Args:
        libraries: A list of `refspy.library.Library` objects to be indexed.
        include_two_letter_aliases: Allow book aliases of only two letters,
            i.e. 'Jn' and '1 Ti'.
        ignore_aliases: A list of alias names to always ignore (e.g. 'Is', 
            since it's ambiguous).
        strict: Throw a ValueError if any two aliases are the same, or any
            alias is an empty string.

    Note:
        Because books like Ruth or Mark have book.abbrev == book.name, we only
        index the abbrev if it differs from the name. These books will not raise
        a ValueError when strict=True.
    """
    index = dict()
    for library in libraries:
        for book in library.books:
            aliases = [book.name]
            if book.name != book.abbrev:
                aliases.append(book.abbrev)
            for alias in book.aliases:
                if alias not in ignore_aliases:
                    len_alias = len(strip_book_number(alias))
                    if (
                        (len_alias > 2 and len_alias < PARAM_CHAR_LIMIT) or
                        (include_two_letter_aliases and len_alias == 2)
                    ):
                        aliases.append(alias)
            for alias in aliases:
                add_unique_book_alias(index, alias, library.id, book.id, strict)
    return index



"""Match biblical references in texts, returning strings and Reference objects."""

import math
import re
from typing import Dict, Generator, List, Match, Tuple

from refspy import language
from refspy.book import Book
from refspy.language import Language
from refspy.range import Range, range, verse_range
from refspy.reference import (
    Reference,
    book_reference,
    reference,
)
from refspy.utils import add_space_after_book_number, get_unnumbered_book_aliases, parse_number, normalize_spacing, trim_trailing_period
from refspy.verse import Number, verse

COLON = r"[:\.]"
DASH = r"[–-]"
END = r"\b"
NUMBER = r"\d+[a-d]?"
SPACE = r"\s*"
RANGE = f"{NUMBER}{DASH}{NUMBER}"
LIST = f"(?:{RANGE}|{NUMBER})(?:\\,\\s*(?:{RANGE}|{NUMBER}))*"

RANGE_OR_NUMBER_COMPILED = re.compile(f"({RANGE}|{NUMBER})")

NUMBER_CAPTURE = re.compile(f"({NUMBER})")
RANGE_CAPTURE = re.compile(f"({NUMBER}){DASH}({NUMBER})")
CHAPTER_RANGE_CAPTURE = re.compile(
    f"{END}({NUMBER}){COLON}({NUMBER}){DASH}({NUMBER}){COLON}({NUMBER}){END}"
)
CHAPTER_VERSES_CAPTURE = re.compile(f"{END}({NUMBER}){COLON}({LIST}){END}")


class Matcher:
    """
    Match:
    - Opening and closing brackets
      - So we don't carry book context past the end of parentheses
    - References
      - Book names (incl. substitutions like 'First' for '1':
        - with no reference: sets context; only returned as a reference if
            requested with include_books=True
        - when book.chapter > 1:
          - 1:2-3:4
          - 1:1,2-3 etc
        - when book.chapter == 1:
          - 1,2-4  (can be verses or chapters, based on book; must not end with
            a number that starts a book name)
      - Stand-alone number references: always have colons or verse markers.
        They rely on context being provided by a previous reference or book
        name.
        - v.1, vv.1-2
        - 1:2-3:4
        - 1:1,2-3 etc (must not end with a number that starts a book name)
    Yield:
      - a (match_str, reference) tuple for each match.
    """

    def __init__(
        self,
        books: Dict[Tuple[Number, Number], Book],
        book_aliases: Dict[str, Tuple[Number, Number]],
        language: Language
    ):
        self.books = books
        self.language = language
        self.book_alias_keys = book_aliases.keys()
        self.book_aliases = self.expand_book_aliases(book_aliases)
        self.brackets_regexp = re.compile(self.build_brackets_regexp())
        self.name_regexp = re.compile(self.build_reference_regexp())

    def build_reference_regexp(self) -> str:
        """
        Reference matches are quadruples. For the string ...

        "Big Book 1:2-5 (cf. Small Book 34) is more interesting than vv.2:3-6."

        ... we want to match:

            ```
            Matching string
            :                 Book Name
            :                  :            Book Reference
            :                  :             :        Numeric Reference
            :                  :             :         :
            ('Big Book 1:2-5', 'Big Book',   '1:2-5',  None)
            ('Small Book 34',  'Small Book', '34',     None)
            ( None,             None,         None,   '2:3-6')
            ```
        """
        NAME_PATTERN = self.build_book_name_regexp()
        VERSE_MARKER = self.build_verse_marker_regexp()
        NUMBER_LIST = self.build_number_list_regexp()
        REGEXP = "".join(
            [
                "(",
                f"{END}({NAME_PATTERN})",  # Rom
                "(",
                "".join(
                    [
                        f"{SPACE}{NUMBER}{COLON}{NUMBER}{DASH}{NUMBER}{COLON}{NUMBER}",  # Rom 1:2-3:4
                        "|",
                        f"{SPACE}{NUMBER}{COLON}{NUMBER_LIST}",  # Rom 3:4,6-9
                        "|",
                        f"{SPACE}{NUMBER_LIST}",  # Phlm 3-4 (verse), Rom 3-4 (chapter)
                    ]
                ),
                ")?",
                ")|(",
                "".join(
                    [
                        f"{NUMBER}{COLON}{NUMBER}{DASH}{NUMBER}{COLON}{NUMBER}",  # 1:2-3:4
                        "|",
                        f"{NUMBER}{COLON}{NUMBER_LIST}",  # 3:4,6-9
                        "|",
                        f"(?:{VERSE_MARKER}){NUMBER_LIST}",  # vv.3-4
                    ]
                ),
                ")",
            ]
        )
        return REGEXP

    def expand_book_aliases(
        self, aliases: Dict[str, Tuple[int, int]]
    ) -> Dict[str, Tuple[int, int]]:
        """
        SBL style requires e.g. "1 Corinthians" be recognized as "First
        Corinthians" at the start of a sentence. Add these as keys to
        the aliases.

        Note:
            Number prefixes are set in the Language class.
        """
        out = {}
        for alias, library_book in aliases.items():
            out[alias] = library_book
            for number, prefixes in self.language.number_prefixes.items():
                if alias.startswith(number + " "):
                    for prefix in prefixes:
                        out[prefix + alias[1:]] = library_book
        return out

    def build_book_name_regexp(self):
        """
        Sort bookname DESC to match the longest first.
        Replace spaces with multi-space matchers in book names.
        Group book names by prefixes.
        Match substitute prefixes for each prefix number:

        Example:
            ```
            (   (1|First|etc)\\s*(Corinthians|etc)
              | (2|Second|etc)\\s*(Corinthians|etc)
              |   ...
              | (Galatians|Romans|etc)
            )
            ```
        """
        regexp_parts = []
        for number in self.numerical_book_prefixes(reverse=True):
            aliases = [
                alias[len(number + " ") :]
                for alias in self.book_alias_keys
                if alias.startswith(number + " ")
            ]
            regexp_parts.append(
                r"(?:"
                + "|".join([number] + self.language.number_prefixes[number])
                + r")"
                + r"\s*"
                + "(?:"
                + "|".join([
                        escape_book_name(_)
                        for _ in long_names_first(aliases)
                    ])
                + ")"
                # + r"(?![^A-Za-z])"  # <-- Need a non-alpha lookahead?
                + r"\b"
                + r"\.?"
            )
        leading_digits = re.compile(r"^\d+\s")
        aliases = [
            alias for alias in self.book_alias_keys if not leading_digits.match(alias)
        ]
        regexp_parts.append(
            r"(?:"
            + "|".join([escape_book_name(_) for _ in long_names_first(aliases)])
            + ")"
            # + r"(?![^A-Za-z])"  # <-- Need a non-alpha lookahead?
            + r"\b"
            + r"\.?"
        )
        return r"|".join(regexp_parts)

    def build_verse_marker_regexp(self):
        return "|".join([re.escape(key) for key in self.language.verse_markers])

    def build_number_list_regexp(self):
        """
        Match a comma-(and-space?)-separated list of numbers and ranges.

        Final numbers in a list require negative lookaheads to ensure they are
        not the start of book names (1 Cor) or chapter refs (6:5):

            ```
            (
                1(?!\\s*(Corinthians|Cor|Thess|etc)))
              | 2(?!\\s*(Corinthians|Cor|Thess|etc)))
              | 3(?!\\s*Jn)
              | 4-999
            )(?!:\\d)
            ```

        TODO:
            Must include missing numbers in the final range.
        """
        regexp_parts = []
        for number in self.numerical_book_prefixes():
            aliases = [
                alias[len(number + " ") :]
                for alias in self.book_aliases.keys()
                if alias.startswith(number + " ")
            ]
            regexp_parts.append(
                number
                + r"(?!"
                + r"\s*"
                + "(?:"  # <-- COLON?
                + "|".join([re.escape(key) for key in sorted(set(aliases))])
                + "))"
            )
        regexp_parts.append(f"[{len(regexp_parts) + 1}-999]")
        SAFE_NUMBER = r"|".join(regexp_parts)
        REGEXP = f"(?:{RANGE}|{NUMBER})(?:\\,\\s*(?:{RANGE}|(?:{SAFE_NUMBER})(?!:\\d){END}))*"
        return REGEXP

    def numerical_book_prefixes(self, reverse=True) -> List[str]:
        prefixes = set()
        for key in self.book_aliases.keys():
            part_1 = key.split(" ")[0]
            if part_1.isdigit():
                prefixes.add(part_1)
        return sorted(prefixes, reverse=reverse)

    def build_brackets_regexp(self) -> str:
        return r"[\(\)]"

    def generate_references(
        self, text: str, include_books: bool = False, include_nones: bool = False
    ) -> Generator[Tuple[str, Reference | None], None, None]:
        """
        Match references and parentheses separately, then take the next lowest
        item (by starting match position) from the regexp match generators, and
        process. Yield references for book names and chapter/verse numbers.

        This is the base function used by `__.find_references()` and
        `__.first_reference()`.

        Args:
            text: In which to find references
            include_books: Whether to match book names alone
            include_nones: Whether to match malformed references
        """
        brackets_matches = self.brackets_regexp.finditer(text)
        reference_matches = self.name_regexp.finditer(text)

        bracket_stack = []

        brackets_match = next(brackets_matches, None)
        reference_match = next(reference_matches, None)

        unnumbered_book_aliases = get_unnumbered_book_aliases(self.book_aliases)

        while reference_match or brackets_match:
            # Handle the start or end of a bracket
            # 0 means bracket, 1 means reference
            values = [
                brackets_match.start() if brackets_match else math.inf,
                reference_match.start() if reference_match else math.inf,
            ]

            book_ref = None
            index_of_minimum = values.index(min(values))

            if brackets_match and index_of_minimum == 0:
                if brackets_match.group(0) == "(":
                    if len(bracket_stack) > 0:
                        bracket_stack.append(bracket_stack[-1])
                if brackets_match.group(0) == ")":
                    if len(bracket_stack) > 0:
                        del bracket_stack[-1]
                brackets_match = next(brackets_matches, None)

            if reference_match and index_of_minimum == 1:
                match_str, book_name, match_with_book, match_without_book = (
                    reference_match.groups()
                )
                try:
                    if book_name:
                        respaced_book_name = add_space_after_book_number(
                            normalize_spacing(trim_trailing_period(book_name)),
                            unnumbered_book_aliases,
                            self.language.number_prefixes
                        )

                        if respaced_book_name in self.book_aliases:
                            library_id, book_id = self.book_aliases[respaced_book_name]
                            book = self.books[library_id, book_id]
                            last_range = book_reference(library_id, book_id).last_range()
                            if match_with_book:
                                if matches := match_chapter_range(match_with_book):
                                    # Rom 1:2-3:4
                                    book_ref = make_chapter_range(last_range, matches)
                                    if book_ref is not None or include_nones:
                                        yield (match_str, book_ref)
                                elif matches := match_chapter_verses(match_with_book):
                                    # Rom 3:4,6-9
                                    book_ref = make_chapter_verses(last_range, matches)
                                    if book_ref is not None or include_nones:
                                        yield (match_str, book_ref)
                                elif matches := match_number_ranges(match_with_book):
                                    if book.chapters == 1:
                                        # Phlm 3-4 (verse)
                                        v = verse_range(
                                            last_range.start.library,
                                            last_range.start.book,
                                            1,
                                            1,
                                        )
                                        book_ref = make_number_ranges(v, matches)
                                        if book_ref is not None or include_nones:
                                            yield (match_str, book_ref)
                                    if book.chapters > 1:
                                        # Rom 3-4 (chapter)
                                        book_ref = make_number_ranges(
                                            last_range, matches, as_chapters=True
                                        )
                                        if book_ref is not None or include_nones:
                                            yield (match_str, book_ref)
                                else:
                                    yield (match_str, None)
                            else:  # no associated reference
                                if bracket_stack:
                                    bracket_stack[-1] = last_range
                                else:
                                    bracket_stack.append(last_range)
                                if include_books:
                                    yield (match_str, book_reference(library_id, book_id))
                    elif match_without_book:
                        if bracket_stack:
                            last_range = bracket_stack[-1]
                            library_id, book_id = (
                                last_range.start.library,
                                last_range.start.book,
                            )
                            if matches := match_chapter_range(match_without_book):
                                # 1:2-3:4
                                book_ref = make_chapter_range(last_range, matches)
                                if book_ref is not None or include_nones:
                                    yield (match_without_book, book_ref)
                            elif matches := match_chapter_verses(match_without_book):
                                # 3:4,6-9
                                book_ref = make_chapter_verses(last_range, matches)
                                if book_ref is not None or include_nones:
                                    yield (match_without_book, book_ref)
                            elif matches := match_number_ranges(match_without_book):
                                # v.2, vv.3-4
                                book_ref = make_number_ranges(last_range, matches)
                                if book_ref is not None or include_nones:
                                    yield (match_without_book, book_ref)
                            else:
                                if include_nones:
                                    yield (match_without_book, None)
                        else:
                            if include_nones:
                                yield (match_without_book, None)

                except ValueError:
                    if include_nones:
                        yield (match_str or match_without_book, None)

                if book_ref:
                    last_range = book_ref.ranges[-1]
                    if bracket_stack:
                        bracket_stack[-1] = last_range
                    else:
                        bracket_stack.append(last_range)

                reference_match = next(reference_matches, None)


def match_chapter_range(text) -> Match | None:
    """Match a pair of chapter-and-verse references.

    Example:
        `Rom 1:2-3:4`
    """
    if match := CHAPTER_RANGE_CAPTURE.search(text):
        return match
    return None


def make_chapter_range(last: Range, match: Match) -> Reference | None:
    """Create pair of chapter-and-verse references from a match.

    See `match_chapter_range`.

    Example:
        `Rom 1:2-3:4`
    """
    if last.is_same_book():
        c1, v1, c2, v2 = match.groups()
        return reference(
            range(
                verse(
                    last.start.library,
                    last.start.book,
                    parse_number(c1),
                    parse_number(v1),
                ),
                verse(
                    last.end.library, last.end.book, parse_number(c2), parse_number(v2)
                ),
            )
        )
    return None


def match_chapter_verses(text) -> Match | None:
    """Match a verse list preceded by a chapter marker.

    Example:
        `Rom 3:4,6-9`
    """
    if match := CHAPTER_VERSES_CAPTURE.search(text):
        return match
    return None


def make_chapter_verses(last: Range, match: Match) -> Reference | None:
    """Create a reference from verse list preceded by a chapter marker.

    See `match_chapter_verses`.

    Example:
        `Rom 3:4,6-9`
    """
    if last.is_same_book():
        chapter, numbers = match.groups()
        if range_match := match_number_ranges(numbers):
            last_range = range(
                verse(last.start.library, last.start.book, int(chapter), 1),
                verse(last.end.library, last.end.book, int(chapter), 999),
            )
            if reference := make_number_ranges(last_range, range_match):
                return reference
    return None


def match_number_ranges(text) -> List[str]:
    """Match a list of numbers and number ranges.

    Example:
        `Phlm 1,3-4` (verse), `Rom 1,3-4` (chapter)
    """
    return RANGE_OR_NUMBER_COMPILED.findall(text)


def make_number_ranges(
    last: Range, matches: List[str], as_chapters: bool = False
) -> Reference | None:
    """Create a reference from a list of numbers and number ranges.

    Args:
        last: a verse to which this number list is relative.
        matches: the result of `match_number_ranges()`
        as_chapters: Treat these numbers as chapters rather than verses.

    Example:
        `Phlm 3-4` (verse), `Rom 3-4` (chapter)

    Note:
        If the end number is less than the start number, we check whether it
        might be an abbreviated reference. `123-24` or even `17-8` can be
        interpreted as `123-124` and `17-18` respectively. We test whether the
        second number has fewer digits than the first.
    """
    ranges = []
    for text in matches:
        if range_matches := RANGE_CAPTURE.match(text):
            start, end = range_matches.group(1, 2)
        elif number_matches := NUMBER_CAPTURE.match(text):
            start = end = number_matches.group(1)
        else:
            continue
        if parse_number(end) < parse_number(start):
            if new_end := infer_abbreviation(start, end):
                end = new_end
            else:
                return None
        if as_chapters:
            if last.is_same_book():
                ranges.append(
                    range(
                        verse(
                            last.start.library, last.start.book, parse_number(start), 1
                        ),
                        verse(last.end.library, last.end.book, parse_number(end), 999),
                    )
                )
        else:
            if last.is_same_chapter():
                ranges.append(
                    range(
                        verse(
                            last.start.library,
                            last.start.book,
                            last.start.chapter,
                            parse_number(start),
                        ),
                        verse(
                            last.end.library,
                            last.end.book,
                            last.end.chapter,
                            parse_number(end),
                        ),
                    )
                )
    if ranges:
        return reference(*ranges)
    else:
        return None


def escape_book_name(name: str) -> str:
    """Regexp escape a name, but match multiple spaces.

    This covers line wrapping, indentation, and so on, in the middle of a book
    name.
    """
    return r"\s+".join([re.escape(_) for _ in name.split(" ")])


def long_names_first(names: List[str]) -> List[str]:
    """Remove duplicates and sort descending."""
    return sorted(set(names), key=len)[::-1]


def infer_abbreviation(start: str, end: str) -> str | None:
    """If a number range ends with a smaller number, try to infer an abbreviation.

    If the end number has fewer digits the the start number then we return a
    copy of the start number with its final digits overprinted by the end
    number.

    * In `123-24`, `24` clearly means `124`.
    * In `12-4`, `4` should probably be interpreted as `14`.
    * In `9-4`, `4` is simply a wrong number. We could guess `14`, but it's too
        uncertain.

    Returns:
        None: if no inference is made.
    """
    if int(end) < int(start):
        if len(end) < len(start):
            return start[: 0 - len(end)] + end
        else:
            return None
    return end



"""A Python package for working with biblical references in ordinary text.

See the [README.md](https://github.com/eukras/refspy) for a more accessible
introduction.
"""

from refspy.init import get_canon, get_language
from refspy.manager import Manager


def refspy(
        canon_name: str = "protestant", 
        locale_name: str = "en_US",
        include_two_letter_aliases: bool = True,
        include_ambiguous_aliases: bool = False
    ) -> Manager:
    """Create a Manager object to access all common package functions.

    See: `refspy.manager.Manager`

    Args:
        canon_name: A valid key for the `refspy.config.LIBRARIES` dict (3)
            - `protestant`
            - `catholic` (adds Deuterocanonicals)
            - `orthodox` (adds Anagignoskomena)
        locale_name: A valid key for the `refspy.config.LANGUAGES` dict (1).
            - `en_US`
        include_two_letter_aliases: e.g. 'Ge', '1 Jn'.
        include_ambiguous_aliases: e.g. 'Am', 'Is'. (See `refspy.languages.*`)

    Note:
        Libraries and languages can be created outside the package, and
        supplied to the Manager object directly. These can be contributed to
        the library, but the point of the library is to read ordinary text
        using ordinary referencing conventions. This will have to be confirmed
        for each proposed library and language.
    """
    return Manager(
        get_canon(canon_name, locale_name),
        get_language(locale_name[:2]),
        include_two_letter_aliases=include_two_letter_aliases,
        include_ambiguous_aliases=include_ambiguous_aliases
    )



from context import *

from pydantic import ValidationError
import pytest

from refspy.reference import (
    chapter_reference,
    reference,
    verse_reference,
)
from refspy.range import range
from refspy.verse import verse


def test_empty_reference_raises_value_error():
    """
    Check empty range lists
    """
    with pytest.raises(ValidationError):
        _ = reference()


def test_reference_comparison():
    reference_1 = reference(range(verse(1, 2, 3, 4), verse(1, 2, 3, 6)))
    reference_2 = reference(range(verse(1, 2, 3, 7), verse(1, 2, 3, 8)))
    assert reference_1 < reference_2
    assert min([reference_1, reference_2]) == reference_1
    assert max([reference_1, reference_2]) == reference_2


def test_reference_addition():
    """
    Check that references add correctly
    """
    range_1 = range(verse(1, 2, 3, 4), verse(1, 2, 3, 6))
    range_2 = range(verse(1, 2, 3, 7), verse(1, 2, 3, 8))
    ref_1 = reference(range_1)
    ref_2 = reference(range_2)
    ref_3 = reference(range_1, range_2)

    assert ref_1 + ref_2 == ref_3


def test_contains():
    ch3 = chapter_reference(1, 2, 3)
    ch3v45 = verse_reference(1, 2, 3, 4, 5)
    assert ch3.contains(ch3)
    assert ch3v45.contains(ch3v45)
    assert ch3.contains(ch3v45)
    assert not ch3v45.contains(ch3)


def test_overlaps():
    ch3 = chapter_reference(1, 2, 3)
    ch3v45 = verse_reference(1, 2, 3, 4, 5)
    assert ch3.overlaps(ch3)
    assert ch3v45.overlaps(ch3v45)
    assert ch3.overlaps(ch3v45)
    assert ch3v45.overlaps(ch3)


def test_adjoins():
    ch3v45 = verse_reference(1, 2, 3, 4, 5)
    ch3v67 = verse_reference(1, 2, 3, 6, 7)
    assert ch3v45.adjoins(ch3v67)
    assert ch3v67.adjoins(ch3v45)



"""Data object for references.

References consist of lists of ranges and are entirely numerical objects.

They can be set to sort, merge, and join references when added together.
"""

from typing import Any, List, Self

from pydantic import BaseModel, Field

from refspy.range import Range, combine, merge, range
from refspy.verse import Number, Verse, verse


class Reference(BaseModel):
    """A reference object represents a list of verse ranges.

    References are entirely numeric entities. Matchers are used to find them in
    text, and formatters are used to turn them in to canonical links.

    It is common to want references to be sorted and combined. Combining means
    merging overlapping ranges and joining adjacent ones. This is done by creating
    references with sorted(), and the `merge`, and `combine` functions in
    `refspy.range.Range`, or calling the methods of the same name on references.

    Example:
        ```
        from reference import merge, combine

        assert reference(*sorted(ranges)) == reference(*ranges).sort()
        assert reference(*merge(ranges)) == reference(*ranges).merge()
        assert reference(*combine(ranges)) == reference(*ranges).combine()
        ```
    """

    ranges: List[Range] = Field(min_length=1)
    """
    A reference must contain at least one `refspy.range.Range`.

    Raises:
        ValueError: If ranges is empty
    """

    def __add__(self, other: Self) -> Self:
        """Overload the addition operator to combine reference ranges into a new object.

        The most aggressive sort/merge/join settings from the two References will
        be used in the new Reference.
        """
        return self.__class__(ranges=[*self.ranges, *other.ranges])

    def __lt__(self, other: Self) -> bool:
        """
        A simple implementation of '<' allows sorting and min/max.

        Note:
            Consider sorting first verse ASC and last verse DESC.
        """
        return self.ranges[0].start < other.ranges[0].start

    def equals(self, other: Self) -> bool:
        """
        Reference equality means that two references have identical ranges.

        Note:
            Consider comparing after any differences in sort/merge/join have
            been reconciled.

        Note:
            We don't use __eq__, as that is already defined in BaseModel.
        """
        return self.ranges == other.ranges

    def overlaps(self, other: Self) -> bool:
        """
        Two references overlap if ANY of their ranges overlap.
        """
        return any(
            [
                self_range.overlaps(other_range)
                for other_range in other.ranges
                for self_range in self.ranges
            ]
        )

    def contains(self, other: Self) -> bool:
        """
        A reference contains another if ALL the other's ranges are contained by
        ANY of it's own ranges.
        """
        return all(
            [
                any([self_range.contains(other_range) for self_range in self.ranges])
                for other_range in other.ranges
            ]
        )

    def adjoins(self, other: Self) -> bool:
        """Determine adjacency for ranges.

        A reference adjoins another if its maximum range adjoins the other's
        minimum or vice-versa. This usually only makes sense for simple
        references with a small number of ranges.

        See `refspy.reference.Reference.adjoins`.
        """
        return any(
            [
                max(self.ranges).adjoins(min(other.ranges)),
                min(self.ranges).adjoins(max(other.ranges)),
            ]
        )

    def count_books(self) -> int:
        """Return the total unique books contained in this reference.

        Allow that the same book ID could appear in multiple libraries.
        """
        library_books = set()
        for _ in self.ranges:
            library_books.add(tuple([_.start.library, _.start.book]))
            library_books.add(tuple([_.end.library, _.end.book]))
        return len(library_books)

    def is_book(self: Self) -> bool:
        """Determine if reference is a whole book."""
        return self.ranges[0].is_book()

    def is_chapter(self: Self) -> bool:
        """Determine if reference is a whole chapter."""
        return self.ranges[0].is_chapter()

    def first_verse(self) -> Verse:
        """Find the first verse.

        Only sensible if sorted=True

        Note:
            Consider sorting on the fly if sorted=False
        """
        return self.ranges[0].start

    def last_verse(self) -> Verse:
        """Find the last verse.

        Only sensible if sorted=True

        Note:
            Consider sorting on the fly if sorted=False
        """
        return self.ranges[-1].end

    def last_range(self) -> Range:
        """Find the last verse.

        Only sensible if sorted=True

        Note:
            Consider sorting on the fly if sorted=False
        """
        return self.ranges[-1]

    def sort(self) -> Self:
        """Return a sorted reference."""
        return self.__class__(ranges=sorted(self.ranges))

    def merge(self) -> Self:
        """Return a merged reference.

        A merged reference is sorted, and has any overlapping ranges merged.
        """
        return self.__class__(ranges=merge(self.ranges))

    def combine(self) -> Self:
        """Return a combined reference.

        A combined reference is sorted, merged, and has any adjacent ranges combined.
        """
        return self.__class__(ranges=combine(self.ranges))


def reference(*args: Range, **kwargs: Any) -> Reference:
    """
    Construct a Reference object from `refspy.range.Range` arguments.

    Example:
        ```
        ref = reference(
            range(verse(1, 2, 3, 3), verse(1, 2, 3, 4)),
            range(verse(1, 2, 3, 6), verse(1, 2, 3, 7))
        )
        ```
    """
    return Reference(ranges=list(args), **kwargs)


def book_reference(library_id: Number, book_id: Number) -> Reference:
    """
    Shorthand function for creating book references from `refspy.number.Number`
    values.
    """
    return reference(
        range(
            verse(library_id, book_id, 1, 1),
            verse(library_id, book_id, 999, 999),
        )
    )


def chapter_reference(
    library_id: Number, book_id: Number, chapter_id: Number
) -> Reference:
    """
    Shorthand function for creating chapter references from
    `refspy.number.Number` values.
    """
    return reference(
        range(
            verse(library_id, book_id, chapter_id, 1),
            verse(library_id, book_id, chapter_id, 999),
        )
    )


def verse_reference(
    library_id: Number,
    book_id: Number,
    chapter_id: Number,
    verse_id: Number,
    verse_end_id: Number | None = None,
) -> Reference:
    """
    Shorthand function for creating verse or range references from
    `refspy.number.Number` values.

    See `refspy.manager.Manager.bcv`
    """
    return reference(
        range(
            verse(library_id, book_id, chapter_id, verse_id),
            verse(library_id, book_id, chapter_id, verse_end_id or verse_id),
        )
    )



from refspy.language import Language


ENGLISH = Language(
    verse_markers=["v.", "vv."],
    ambiguous_aliases=['Am', 'Ho', 'Is', 'So'],
    number_prefixes={
        "1": ["First", "1st", "I"],
        "2": ["Second", "2nd", "II"],
        "3": ["Third", "3rd", "III"],
        "4": ["Fourth", "4th", "IV"],
    },
)



from context import *

from refspy import refspy
from refspy.libraries.en_US import NT
from refspy.reference import verse_reference

__ = refspy()

def test_numbered_books():
    """
    Simple sanity checks for the books in the en_US libraries.
    """
    assert __.r('1 Cor 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('2 Cor 1:4') == verse_reference(NT.id, 8, 1, 4)
    assert __.r('1 Corinthians 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('2 Corinthians 1:4') == verse_reference(NT.id, 8, 1, 4)

    assert __.r('First Cor 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('First Corinthians 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('Second Cor 1:4') == verse_reference(NT.id, 8, 1, 4)
    assert __.r('Second Corinthians 1:4') == verse_reference(NT.id, 8, 1, 4)

    assert __.r('1st Cor 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('1st Corinthians 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('2nd Cor 1:4') == verse_reference(NT.id, 8, 1, 4)
    assert __.r('2nd Corinthians 1:4') == verse_reference(NT.id, 8, 1, 4)

    assert __.r('I Cor 1:4') == verse_reference(NT.id, 7, 1, 4)
    assert __.r('I Corinthians 1:4') == verse_reference(NT.id, 7, 1, 4)

    assert __.r('II Cor 1:4') == verse_reference(NT.id, 8, 1, 4)
    assert __.r('II Corinthians 1:4') == verse_reference(NT.id, 8, 1, 4)



from refspy.book import Book
from refspy.library import Library

# See README.md for Library ID numbers
# English language libraries follow SBL style guide for books and
# abbreviations.

OT = Library(
    id=200,
    name="Old Testament",
    abbrev="OT",
    books=[
        Book(
            id=1,
            name="Genesis",
            abbrev="Gen",
            aliases=["Ge", "Gn"],
            chapters=50,
        ),
        Book(
            id=2,
            name="Exodus",
            abbrev="Exod",
            aliases=["Exo", "Ex"],
            chapters=40,
        ),
        Book(
            id=3,
            name="Leviticus",
            abbrev="Lev",
            aliases=["Le"],
            chapters=27,
        ),
        Book(
            id=4,
            name="Numbers",
            abbrev="Num",
            aliases=["Nu"],
            chapters=36,
        ),
        Book(
            id=5,
            name="Deuteronomy",
            abbrev="Deut",
            aliases=["De", "Dt"],
            chapters=34,
        ),
        Book(
            id=6,
            name="Joshua",
            abbrev="Josh",
            aliases=["Jos"],
            chapters=24,
        ),
        Book(
            id=7,
            name="Judges",
            abbrev="Judg",
            aliases=["Jdg"],
            chapters=21,
        ),
        Book(
            id=8,
            name="Ruth",
            abbrev="Ruth",
            aliases=["Rut", "Ru"],
            chapters=4,
        ),
        Book(
            id=9,
            name="1 Samuel",
            abbrev="1 Sam",
            aliases=["1 Sa"],
            chapters=31,
        ),
        Book(
            id=10,
            name="2 Samuel",
            abbrev="2 Sam",
            aliases=["2 Sa"],
            chapters=24,
        ),
        Book(
            id=11,
            name="1 Kings",
            abbrev="1 Kgs",
            aliases=["1 Ki"],
            chapters=22,
        ),
        Book(
            id=12,
            name="2 Kings",
            abbrev="2 Kgs",
            aliases=["2 Ki"],
            chapters=22,
        ),
        Book(
            id=13,
            name="1 Chronicles",
            abbrev="1 Chr",
            aliases=["1 Ch"],
            chapters=29,
        ),
        Book(
            id=14,
            name="2 Chronicles",
            abbrev="2 Chr",
            aliases=["2 Ch"],
            chapters=36,
        ),
        Book(
            id=15,
            name="Ezra",
            abbrev="Ezra",
            aliases=["Ezr"],
            chapters=10,
        ),
        Book(
            id=16,
            name="Nehemiah",
            abbrev="Neh",
            aliases=["Ne"],
            chapters=13,
        ),
        Book(
            id=17,
            name="Esther",
            abbrev="Esth",
            aliases=["Est", "Es"],
            chapters=10,
        ),
        Book(
            id=18,
            name="Job",
            abbrev="Job",
            aliases=[],
            chapters=42,
        ),
        Book(
            id=19,
            name="Psalm",
            abbrev="Ps",
            aliases=["Psa"],
            chapters=150,
        ),
        Book(
            id=20,
            name="Proverbs",
            abbrev="Prov",
            aliases=["Pro", "Pr"],
            chapters=31,
        ),
        Book(
            id=21,
            name="Ecclesiastes",
            abbrev="Eccl",
            aliases=["Qoheleth", "Qoh", "Ecc", "Ec"],
            chapters=12,
        ),
        Book(
            id=22,
            name="Song of Solomon",
            abbrev="Song",
            aliases=["Canticles", "Cant", "Sng", "So"],
            chapters=8,
        ),
        Book(
            id=23,
            name="Isaiah",
            abbrev="Isa",
            aliases=["Is"],
            chapters=66,
        ),
        Book(
            id=24,
            name="Jeremiah",
            abbrev="Jer",
            aliases=["Je"],
            chapters=52,
        ),
        Book(
            id=25,
            name="Lamentations",
            abbrev="Lam",
            aliases=["La"],
            chapters=5,
        ),
        Book(
            id=26,
            name="Ezekiel",
            abbrev="Ezek",
            aliases=["Ezk", "Eze"],
            chapters=48,
        ),
        Book(
            id=27,
            name="Daniel",
            abbrev="Dan",
            aliases=["Da"],
            chapters=12,
        ),
        Book(
            id=28,
            name="Hosea",
            abbrev="Hos",
            aliases=["Ho"],
            chapters=14,
        ),
        Book(
            id=29,
            name="Joel",
            abbrev="Joel",
            aliases=["Joe", "Jol"],
            chapters=3,
        ),
        Book(
            id=30,
            name="Amos",
            abbrev="Amos",
            aliases=["Amo", "Am"],
            chapters=9,
        ),
        Book(
            id=31,
            name="Obadiah",
            abbrev="Obad",
            aliases=["Oba", "Ob"],
            chapters=1,
        ),
        Book(
            id=32,
            name="Jonah",
            abbrev="Jonah",
            aliases=["Jon"],
            chapters=4,
        ),
        Book(
            id=33,
            name="Micah",
            abbrev="Mic",
            aliases=[],
            chapters=7,
        ),
        Book(
            id=34,
            name="Nahum",
            abbrev="Nah",
            aliases=["Na"],
            chapters=3,
        ),
        Book(
            id=35,
            name="Habbakuk",
            abbrev="Hab",
            aliases=[],
            chapters=3,
        ),
        Book(
            id=36,
            name="Zephaniah",
            abbrev="Zeph",
            aliases=["Zep"],
            chapters=3,
        ),
        Book(
            id=37,
            name="Haggai",
            abbrev="Hag",
            aliases=["Hg"],
            chapters=2,
        ),
        Book(
            id=38,
            name="Zechariah",
            abbrev="Zech",
            aliases=["Zec"],
            chapters=14,
        ),
        Book(
            id=39,
            name="Malachi",
            abbrev="Mal",
            aliases=[],
            chapters=4,
        ),
    ],
)

DC = Library(
    id=210,
    name="Deuterocanonical",
    abbrev="DC",
    books=[
        Book(
            id=1,
            name="Tobit",
            abbrev="Tob",
            aliases=[],
            chapters=14,
        ),
        Book(
            id=2,
            name="Judith",
            abbrev="Jdt",
            aliases=[],
            chapters=16,
        ),
        Book(
            # TODO: Offsets: 11..16? 
            id=3,
            name="Additions to Esther",
            abbrev="Add Esth",
            aliases=["Esg"], # "Esther Greek"
            chapters=10,
        ),
        Book(
            id=4,
            name="Wisdom of Solomon",
            abbrev="Wis",
            aliases=[],
            chapters=19,
        ),
        Book(
            id=5,
            name="Sirach",
            abbrev="Sir",
            aliases=["Ecclesiasticus", "Eccles"],
            chapters=51,
        ),
        Book(
            id=6,
            name="Baruch",
            abbrev="Bar",
            aliases=[],
            chapters=6,
        ),
        Book(
            id=7,
            name="Epistle of Jeremiah",
            abbrev="Ep Jer",
            aliases=["Letter of Jeremiah", "LJe"],
            chapters=1,
        ),
        Book(
            id=9,
            name="Prayer of Azariah",
            abbrev="Pr Azar",
            aliases=[],
            chapters=1,
        ),
        Book(
            id=10,
            name="Song of the Three Young Men",
            abbrev="Sg Three",
            aliases=["Song of the Three Youths", "S3Y"],
            chapters=1,
        ),
        Book(
            id=11,
            name="Susannah",
            abbrev="Sus",
            aliases=[],
            chapters=1,
        ),
        Book(
            id=12,
            name="Bel and the Dragon",
            abbrev="Bel",
            aliases=[],
            chapters=1,
        ),
        Book(
            id=13,
            name="1 Maccabees",
            abbrev="1 Macc",
            aliases=["1 Mac", "1 Ma"],
            chapters=16,
        ),
        Book(
            id=14,
            name="2 Maccabees",
            abbrev="2 Macc",
            aliases=["2 Mac", "2 Ma"],
            chapters=15,
        ),
    ],
)
DC_ORTHODOX = Library(
    id=220,
    name="Deuterocanonical (Orthodox)",
    abbrev="DCO",
    books=[
        Book(
            id=1,
            name="1 Esdras",
            abbrev="1 Esd",
            aliases=["1 Es"],
            chapters=9,
        ),
        Book(
            id=2,
            name="Prayer of Manesseh",
            abbrev="Pr Man",
            aliases=[],
            chapters=1,
        ),
        Book(
            id=3,
            name="Psalm 151",
            abbrev="Ps 151",
            aliases=["Add Ps", "AddPs"],
            chapters=1,
        ),
        Book(
            id=4,
            name="3 Maccabees",
            abbrev="3 Macc",
            aliases=["3 Mac", "3 Ma"],
            chapters=7,
        ),
        Book(
            id=5,
            name="2 Esdras",
            abbrev="2 Esd",
            aliases=["2 Es"],
            chapters=16,
        ),
        Book(
            id=6,
            name="4 Maccabees",
            abbrev="4 Macc",
            aliases=["4 Mac", "4 Ma"],
            chapters=18,
        ),
    ],
)

NT = Library(
    id=400,
    name="New Testament",
    abbrev="NT",
    books=[
        Book(
            id=1,
            name="Matthew",
            abbrev="Matt",
            aliases=["Mat", "Mt"],
            chapters=28,
        ),
        Book(
            id=2,
            name="Mark",
            abbrev="Mark",
            aliases=["Mrk", "Mk"],
            chapters=16,
        ),
        Book(
            id=3,
            name="Luke",
            abbrev="Luke",
            aliases=["Luk", "Lk"],
            chapters=24,
        ),
        Book(
            id=4,
            name="John",
            abbrev="John",
            aliases=["Jhn", "Jn"],
            chapters=21,
        ),
        Book(
            id=5,
            name="Acts",
            abbrev="Acts",
            aliases=["Act", "Ac"],
            chapters=28,
        ),
        Book(
            id=6,
            name="Romans",
            abbrev="Rom",
            aliases=["Ro"],
            chapters=16,
        ),
        Book(
            id=7,
            name="1 Corinthians",
            abbrev="1 Cor",
            aliases=["1 Co"],
            chapters=16,
        ),
        Book(
            id=8,
            name="2 Corinthians",
            abbrev="2 Cor",
            aliases=["2 Co"],
            chapters=13,
        ),
        Book(
            id=9,
            name="Galatians",
            abbrev="Gal",
            aliases=["Ga"],
            chapters=6,
        ),
        Book(
            id=10,
            name="Ephesians",
            abbrev="Eph",
            aliases=[],
            chapters=6,
        ),
        Book(
            id=11,
            name="Philippians",
            abbrev="Phil",
            aliases=["Php", "Phlp"],
            chapters=4,
        ),
        Book(
            id=12,
            name="Colossians",
            abbrev="Col",
            aliases=[],
            chapters=4,
        ),
        Book(
            id=13,
            name="1 Thessalonians",
            abbrev="1 Thess",
            aliases=["1 Th"],
            chapters=5,
        ),
        Book(
            id=14,
            name="2 Thessalonians",
            abbrev="2 Thess",
            aliases=["2 Th"],
            chapters=3,
        ),
        Book(
            id=15,
            name="1 Timothy",
            abbrev="1 Tim",
            aliases=["1 Ti"],
            chapters=6,
        ),
        Book(
            id=16,
            name="2 Timothy",
            abbrev="2 Tim",
            aliases=["2 Ti"],
            chapters=4,
        ),
        Book(
            id=17,
            name="Titus",
            abbrev="Tit",
            aliases=["Tt"],
            chapters=3,
        ),
        Book(
            id=18,
            name="Philemon",
            abbrev="Phlm",
            aliases=["Phm"],
            chapters=1,
        ),
        Book(
            id=19,
            name="Hebrews",
            abbrev="Heb",
            aliases=[],
            chapters=13,
        ),
        Book(
            id=20,
            name="James",
            abbrev="Jam",
            aliases=["Jas"],
            chapters=5,
        ),
        Book(
            id=21,
            name="1 Peter",
            abbrev="1 Pet",
            aliases=["1 Pe"],
            chapters=5,
        ),
        Book(
            id=22,
            name="2 Peter",
            abbrev="2 Pet",
            aliases=["2 Pe"],
            chapters=3,
        ),
        Book(
            id=23,
            name="1 John",
            abbrev="1 John",
            aliases=["1 Jn"],
            chapters=5,
        ),
        Book(
            id=24,
            name="2 John",
            abbrev="2 John",
            aliases=["2 Jn"],
            chapters=1,
        ),
        Book(
            id=25,
            name="3 John",
            abbrev="3 John",
            aliases=["3 Jn"],
            chapters=1,
        ),
        Book(
            id=26,
            name="Jude",
            abbrev="Jude",
            aliases=["Jd"],
            chapters=1,
        ),
        Book(
            id=27,
            name="Revelation",
            abbrev="Rev",
            aliases=["Apocalypse", "Apoc", "Re", "Rv"], 
            chapters=22,
        ),
    ],
)
